import time
import typing as tp

from anyio import to_thread
from botocore.exceptions import ClientError


def get_timestamp_in_ms() -> float:
    return time.time() * 1000


class S3Manager:
    def __init__(
        self,
        client: tp.Any,
        bucket_name: str,
        check_ttl_every: tp.Union[int, float],
        is_binary: bool = False,
        path_prefix: str = "hishel-",
    ):
        self._client = client
        self._bucket_name = bucket_name
        self._is_binary = is_binary
        self._last_cleaned = time.monotonic()
        self._check_ttl_every = check_ttl_every
        self._path_prefix = path_prefix

    def write_to(self, path: str, data: tp.Union[bytes, str], only_metadata: bool = False) -> None:
        path = self._path_prefix + path
        if isinstance(data, str):
            data = data.encode("utf-8")

        created_at = None
        if only_metadata:
            try:
                response = self._client.get_object(
                    Bucket=self._bucket_name,
                    Key=path,
                )
                created_at = response["Metadata"]["created_at"]
            except Exception:
                pass

        self._client.put_object(
            Bucket=self._bucket_name,
            Key=path,
            Body=data,
            Metadata={"created_at": created_at or str(get_timestamp_in_ms())},
        )

    def read_from(self, path: str) -> tp.Union[bytes, str]:
        path = self._path_prefix + path
        response = self._client.get_object(
            Bucket=self._bucket_name,
            Key=path,
        )

        content = response["Body"].read()

        if self._is_binary:  # pragma: no cover
            return tp.cast(bytes, content)

        return tp.cast(str, content.decode("utf-8"))

    def remove_expired(self, ttl: int, key: str) -> None:
        path = self._path_prefix + key

        if time.monotonic() - self._last_cleaned < self._check_ttl_every:
            try:
                response = self._client.get_object(Bucket=self._bucket_name, Key=path)
                if get_timestamp_in_ms() - float(response["Metadata"]["created_at"]) > ttl:
                    self._client.delete_object(Bucket=self._bucket_name, Key=path)
                return
            except ClientError as e:
                if e.response["Error"]["Code"] == "NoSuchKey":
                    return
                raise e

        self._last_cleaned = time.monotonic()
        for obj in self._client.list_objects(Bucket=self._bucket_name).get("Contents", []):
            if not obj["Key"].startswith(self._path_prefix):  # pragma: no cover
                continue

            try:
                metadata_obj = self._client.head_object(Bucket=self._bucket_name, Key=obj["Key"]).get("Metadata", {})
            except ClientError as e:
                if e.response["Error"]["Code"] == "404":
                    continue

            if not metadata_obj or "created_at" not in metadata_obj:
                continue

            if get_timestamp_in_ms() - float(metadata_obj["created_at"]) > ttl:
                self._client.delete_object(Bucket=self._bucket_name, Key=obj["Key"])

    def remove_entry(self, key: str) -> None:
        path = self._path_prefix + key
        self._client.delete_object(Bucket=self._bucket_name, Key=path)


class AsyncS3Manager:  # pragma: no cover
    def __init__(
        self,
        client: tp.Any,
        bucket_name: str,
        check_ttl_every: tp.Union[int, float],
        is_binary: bool = False,
        path_prefix: str = "hishel-",
    ):
        self._sync_manager = S3Manager(client, bucket_name, check_ttl_every, is_binary, path_prefix)

    async def write_to(self, path: str, data: tp.Union[bytes, str], only_metadata: bool = False) -> None:
        return await to_thread.run_sync(self._sync_manager.write_to, path, data, only_metadata)

    async def read_from(self, path: str) -> tp.Union[bytes, str]:
        return await to_thread.run_sync(self._sync_manager.read_from, path)

    async def remove_expired(self, ttl: int, key: str) -> None:
        return await to_thread.run_sync(self._sync_manager.remove_expired, ttl, key)

    async def remove_entry(self, key: str) -> None:
        return await to_thread.run_sync(self._sync_manager.remove_entry, key)



import typing as tp

import httpx

from hishel._sync._transports import CacheTransport

__all__ = ("CacheClient",)


class CacheClient(httpx.Client):
    def __init__(self, *args: tp.Any, **kwargs: tp.Any):
        self._storage = kwargs.pop("storage") if "storage" in kwargs else None
        self._controller = kwargs.pop("controller") if "controller" in kwargs else None
        super().__init__(*args, **kwargs)

    def _init_transport(self, *args, **kwargs) -> CacheTransport:  # type: ignore
        _transport = super()._init_transport(*args, **kwargs)
        return CacheTransport(
            transport=_transport,
            storage=self._storage,
            controller=self._controller,
        )

    def _init_proxy_transport(self, *args, **kwargs) -> CacheTransport:  # type: ignore
        _transport = super()._init_proxy_transport(*args, **kwargs)  # pragma: no cover
        return CacheTransport(  # pragma: no cover
            transport=_transport,
            storage=self._storage,
            controller=self._controller,
        )



import string
from typing import Any, Dict, List, Optional, Union

from ._exceptions import ParseError, ValidationError

## Grammar


HTAB = "\t"
SP = " "
obs_text = "".join(chr(i) for i in range(0x80, 0xFF + 1))  # 0x80-0xFF

tchar = "!#$%&'*+-.^_`|~0123456789" + string.ascii_letters
qdtext = "".join(
    [
        HTAB,
        SP,
        "\x21",
        "".join(chr(i) for i in range(0x23, 0x5B + 1)),  # 0x23-0x5b
        "".join(chr(i) for i in range(0x5D, 0x7E + 1)),  # 0x5D-0x7E
        obs_text,
    ]
)

TIME_FIELDS = [
    "max_age",
    "max_stale",
    "min_fresh",
    "s_maxage",
]

BOOLEAN_FIELDS = [
    "immutable",
    "must_revalidate",
    "must_understand",
    "no_store",
    "no_transform",
    "only_if_cached",
    "public",
    "proxy_revalidate",
]

LIST_FIELDS = ["no_cache", "private"]

__all__ = (
    "CacheControl",
    "Vary",
)


def strip_ows_around(text: str) -> str:
    return text.strip(" ").strip("\t")


def normalize_directive(text: str) -> str:
    return text.replace("-", "_")


def parse_cache_control(cache_control_values: List[str]) -> "CacheControl":
    directives = {}

    for cache_control_value in cache_control_values:
        if "no-cache=" in cache_control_value or "private=" in cache_control_value:
            cache_control_splited = [cache_control_value]
        else:
            cache_control_splited = cache_control_value.split(",")

        for directive in cache_control_splited:
            key: str = ""
            value: Optional[str] = None
            dquote = False

            if not directive:
                raise ParseError("The directive should not be left blank.")

            directive = strip_ows_around(directive)

            if not directive:
                raise ParseError("The directive should not contain only whitespaces.")

            for i, key_char in enumerate(directive):
                if key_char == "=":
                    value = directive[i + 1 :]

                    if not value:
                        raise ParseError("The directive value cannot be left blank.")

                    if value[0] == '"':
                        dquote = True
                    if dquote and value[-1] != '"':
                        raise ParseError("Invalid quotes around the value.")

                    if not dquote:
                        for value_char in value:
                            if value_char not in tchar:
                                raise ParseError(
                                    f"The character '{value_char!r}' is not permitted for the unquoted values."
                                )
                    else:
                        for value_char in value[1:-1]:
                            if value_char not in qdtext:
                                raise ParseError(
                                    f"The character '{value_char!r}' is not permitted for the quoted values."
                                )
                    break

                if key_char not in tchar:
                    raise ParseError(f"The character '{key_char!r}' is not permitted in the directive name.")
                key += key_char
            directives[key] = value
    validated_data = CacheControl.validate(directives)
    return CacheControl(**validated_data)


class Vary:
    def __init__(self, values: List[str]) -> None:
        self._values = values

    @classmethod
    def from_value(cls, vary_values: List[str]) -> "Vary":
        values = []

        for vary_value in vary_values:
            for field_name in vary_value.split(","):
                field_name = field_name.strip()
                values.append(field_name)
        return Vary(values)


class CacheControl:
    def __init__(
        self,
        immutable: bool = False,  # [RFC8246]
        max_age: Optional[int] = None,  # [RFC9111, Section 5.2.1.1, 5.2.2.1]
        max_stale: Optional[int] = None,  # [RFC9111, Section 5.2.1.2]
        min_fresh: Optional[int] = None,  # [RFC9111, Section 5.2.1.3]
        must_revalidate: bool = False,  # [RFC9111, Section 5.2.2.2]
        must_understand: bool = False,  # [RFC9111, Section 5.2.2.3]
        no_cache: Union[bool, List[str]] = False,  # [RFC9111, Section 5.2.1.4, 5.2.2.4]
        no_store: bool = False,  # [RFC9111, Section 5.2.1.5, 5.2.2.5]
        no_transform: bool = False,  # [RFC9111, Section 5.2.1.6, 5.2.2.6]
        only_if_cached: bool = False,  # [RFC9111, Section 5.2.1.7]
        private: Union[bool, List[str]] = False,  # [RFC9111, Section 5.2.2.7]
        proxy_revalidate: bool = False,  # [RFC9111, Section 5.2.2.8]
        public: bool = False,  # [RFC9111, Section 5.2.2.9]
        s_maxage: Optional[int] = None,  # [RFC9111, Section 5.2.2.10]
    ) -> None:
        self.immutable = immutable
        self.max_age = max_age
        self.max_stale = max_stale
        self.min_fresh = min_fresh
        self.must_revalidate = must_revalidate
        self.must_understand = must_understand
        self.no_cache = no_cache
        self.no_store = no_store
        self.no_transform = no_transform
        self.only_if_cached = only_if_cached
        self.private = private
        self.proxy_revalidate = proxy_revalidate
        self.public = public
        self.s_maxage = s_maxage

    @classmethod
    def validate(cls, directives: Dict[str, Any]) -> Dict[str, Any]:
        validated_data: Dict[str, Any] = {}

        for key, value in directives.items():
            key = normalize_directive(key)
            if key in TIME_FIELDS:
                if value is None:
                    raise ValidationError(f"The directive '{key}' necessitates a value.")

                if value[0] == '"' or value[-1] == '"':
                    raise ValidationError(f"The argument '{key}' should be an integer, but a quote was found.")

                try:
                    validated_data[key] = int(value)
                except Exception:
                    raise ValidationError(f"The argument '{key}' should be an integer, but got '{value!r}'.")
            elif key in BOOLEAN_FIELDS:
                if value is not None:
                    raise ValidationError(f"The directive '{key}' should have no value, but it does.")
                validated_data[key] = True
            elif key in LIST_FIELDS:
                if value is None:
                    validated_data[key] = True
                else:
                    values = []
                    for list_value in value[1:-1].split(","):
                        if not list_value:
                            raise ValidationError("The list value must not be empty.")
                        list_value = strip_ows_around(list_value)
                        values.append(list_value)
                    validated_data[key] = values

        return validated_data

    def __repr__(self) -> str:
        fields = ""

        for key in TIME_FIELDS:
            key = key.replace("-", "_")
            value = getattr(self, key)
            if value:
                fields += f"{key}={value}, "

        for key in BOOLEAN_FIELDS:
            key = key.replace("-", "_")
            value = getattr(self, key)
            if value:
                fields += f"{key}, "

        fields = fields[:-2]

        return f"<{type(self).__name__} {fields}>"



import base64
import json
import pickle
import typing as tp
from datetime import datetime

from httpcore import Request, Response

from hishel._utils import normalized_url

try:
    import yaml
except ImportError:  # pragma: no cover
    yaml = None  # type: ignore

HEADERS_ENCODING = "iso-8859-1"
KNOWN_RESPONSE_EXTENSIONS = ("http_version", "reason_phrase")
KNOWN_REQUEST_EXTENSIONS = ("timeout", "sni_hostname")

__all__ = ("PickleSerializer", "JSONSerializer", "YAMLSerializer", "BaseSerializer", "clone_model")

T = tp.TypeVar("T", Request, Response)


def clone_model(model: T) -> T:
    if isinstance(model, Response):
        return Response(
            status=model.status,
            headers=model.headers,
            content=model.content,
            extensions={key: value for key, value in model.extensions.items() if key in KNOWN_RESPONSE_EXTENSIONS},
        )  # type: ignore
    else:
        return Request(
            method=model.method,
            url=normalized_url(model.url),
            headers=model.headers,
            extensions={key: value for key, value in model.extensions.items() if key in KNOWN_REQUEST_EXTENSIONS},
        )  # type: ignore


class Metadata(tp.TypedDict):
    number_of_uses: int
    created_at: datetime
    cache_key: str


class BaseSerializer:
    def dumps(self, response: Response, request: Request, metadata: Metadata) -> tp.Union[str, bytes]:
        raise NotImplementedError()

    def loads(self, data: tp.Union[str, bytes]) -> tp.Tuple[Response, Request, Metadata]:
        raise NotImplementedError()

    @property
    def is_binary(self) -> bool:
        raise NotImplementedError()


class PickleSerializer(BaseSerializer):
    """
    A simple pickle-based serializer.
    """

    def dumps(self, response: Response, request: Request, metadata: Metadata) -> tp.Union[str, bytes]:
        """
        Dumps the HTTP response and its HTTP request.

        :param response: An HTTP response
        :type response: Response
        :param request: An HTTP request
        :type request: Request
        :param metadata: Additional information about the stored response
        :type metadata: Metadata
        :return: Serialized response
        :rtype: tp.Union[str, bytes]
        """
        clone_response = clone_model(response)
        clone_request = clone_model(request)
        return pickle.dumps((clone_response, clone_request, metadata))

    def loads(self, data: tp.Union[str, bytes]) -> tp.Tuple[Response, Request, Metadata]:
        """
        Loads the HTTP response and its HTTP request from serialized data.

        :param data: Serialized data
        :type data: tp.Union[str, bytes]
        :return: HTTP response and its HTTP request
        :rtype: tp.Tuple[Response, Request, Metadata]
        """
        assert isinstance(data, bytes)
        return tp.cast(tp.Tuple[Response, Request, Metadata], pickle.loads(data))

    @property
    def is_binary(self) -> bool:  # pragma: no cover
        return True


class JSONSerializer(BaseSerializer):
    """A simple json-based serializer."""

    def dumps(self, response: Response, request: Request, metadata: Metadata) -> tp.Union[str, bytes]:
        """
        Dumps the HTTP response and its HTTP request.

        :param response: An HTTP response
        :type response: Response
        :param request: An HTTP request
        :type request: Request
        :param metadata: Additional information about the stored response
        :type metadata: Metadata
        :return: Serialized response
        :rtype: tp.Union[str, bytes]
        """
        response_dict = {
            "status": response.status,
            "headers": [
                (key.decode(HEADERS_ENCODING), value.decode(HEADERS_ENCODING)) for key, value in response.headers
            ],
            "content": base64.b64encode(response.content).decode("ascii"),
            "extensions": {
                key: value.decode("ascii")
                for key, value in response.extensions.items()
                if key in KNOWN_RESPONSE_EXTENSIONS
            },
        }

        request_dict = {
            "method": request.method.decode("ascii"),
            "url": normalized_url(request.url),
            "headers": [
                (key.decode(HEADERS_ENCODING), value.decode(HEADERS_ENCODING)) for key, value in request.headers
            ],
            "extensions": {key: value for key, value in request.extensions.items() if key in KNOWN_REQUEST_EXTENSIONS},
        }

        metadata_dict = {
            "cache_key": metadata["cache_key"],
            "number_of_uses": metadata["number_of_uses"],
            "created_at": metadata["created_at"].strftime("%a, %d %b %Y %H:%M:%S GMT"),
        }

        full_json = {
            "response": response_dict,
            "request": request_dict,
            "metadata": metadata_dict,
        }

        return json.dumps(full_json, indent=4)

    def loads(self, data: tp.Union[str, bytes]) -> tp.Tuple[Response, Request, Metadata]:
        """
        Loads the HTTP response and its HTTP request from serialized data.

        :param data: Serialized data
        :type data: tp.Union[str, bytes]
        :return: HTTP response and its HTTP request
        :rtype: tp.Tuple[Response, Request, Metadata]
        """

        full_json = json.loads(data)

        response_dict = full_json["response"]
        request_dict = full_json["request"]
        metadata_dict = full_json["metadata"]
        metadata_dict["created_at"] = datetime.strptime(
            metadata_dict["created_at"],
            "%a, %d %b %Y %H:%M:%S GMT",
        )

        response = Response(
            status=response_dict["status"],
            headers=[
                (key.encode(HEADERS_ENCODING), value.encode(HEADERS_ENCODING))
                for key, value in response_dict["headers"]
            ],
            content=base64.b64decode(response_dict["content"].encode("ascii")),
            extensions={
                key: value.encode("ascii")
                for key, value in response_dict["extensions"].items()
                if key in KNOWN_RESPONSE_EXTENSIONS
            },
        )

        request = Request(
            method=request_dict["method"],
            url=request_dict["url"],
            headers=[
                (key.encode(HEADERS_ENCODING), value.encode(HEADERS_ENCODING)) for key, value in request_dict["headers"]
            ],
            extensions={
                key: value for key, value in request_dict["extensions"].items() if key in KNOWN_REQUEST_EXTENSIONS
            },
        )

        metadata = Metadata(
            cache_key=metadata_dict["cache_key"],
            created_at=metadata_dict["created_at"],
            number_of_uses=metadata_dict["number_of_uses"],
        )

        return response, request, metadata

    @property
    def is_binary(self) -> bool:
        return False


class YAMLSerializer(BaseSerializer):
    """A simple yaml-based serializer."""

    def dumps(self, response: Response, request: Request, metadata: Metadata) -> tp.Union[str, bytes]:
        """
        Dumps the HTTP response and its HTTP request.

        :param response: An HTTP response
        :type response: Response
        :param request: An HTTP request
        :type request: Request
        :param metadata: Additional information about the stored response
        :type metadata: Metadata
        :return: Serialized response
        :rtype: tp.Union[str, bytes]
        """
        if yaml is None:  # pragma: no cover
            raise RuntimeError(
                f"The `{type(self).__name__}` was used, but the required packages were not found. "
                "Check that you have `Hishel` installed with the `yaml` extension as shown.\n"
                "```pip install hishel[yaml]```"
            )
        response_dict = {
            "status": response.status,
            "headers": [
                (key.decode(HEADERS_ENCODING), value.decode(HEADERS_ENCODING)) for key, value in response.headers
            ],
            "content": base64.b64encode(response.content).decode("ascii"),
            "extensions": {
                key: value.decode("ascii")
                for key, value in response.extensions.items()
                if key in KNOWN_RESPONSE_EXTENSIONS
            },
        }

        request_dict = {
            "method": request.method.decode("ascii"),
            "url": normalized_url(request.url),
            "headers": [
                (key.decode(HEADERS_ENCODING), value.decode(HEADERS_ENCODING)) for key, value in request.headers
            ],
            "extensions": {key: value for key, value in request.extensions.items() if key in KNOWN_REQUEST_EXTENSIONS},
        }

        metadata_dict = {
            "cache_key": metadata["cache_key"],
            "number_of_uses": metadata["number_of_uses"],
            "created_at": metadata["created_at"].strftime("%a, %d %b %Y %H:%M:%S GMT"),
        }

        full_json = {
            "response": response_dict,
            "request": request_dict,
            "metadata": metadata_dict,
        }

        return yaml.safe_dump(full_json, sort_keys=False)

    def loads(self, data: tp.Union[str, bytes]) -> tp.Tuple[Response, Request, Metadata]:
        """
        Loads the HTTP response and its HTTP request from serialized data.

        :param data: Serialized data
        :type data: tp.Union[str, bytes]
        :raises RuntimeError: When used without the `yaml` extension installed
        :return: HTTP response and its HTTP request
        :rtype: tp.Tuple[Response, Request, Metadata]
        """
        if yaml is None:  # pragma: no cover
            raise RuntimeError(
                f"The `{type(self).__name__}` was used, but the required packages were not found. "
                "Check that you have `Hishel` installed with the `yaml` extension as shown.\n"
                "```pip install hishel[yaml]```"
            )

        full_json = yaml.safe_load(data)

        response_dict = full_json["response"]
        request_dict = full_json["request"]
        metadata_dict = full_json["metadata"]
        metadata_dict["created_at"] = datetime.strptime(
            metadata_dict["created_at"],
            "%a, %d %b %Y %H:%M:%S GMT",
        )

        response = Response(
            status=response_dict["status"],
            headers=[
                (key.encode(HEADERS_ENCODING), value.encode(HEADERS_ENCODING))
                for key, value in response_dict["headers"]
            ],
            content=base64.b64decode(response_dict["content"].encode("ascii")),
            extensions={
                key: value.encode("ascii")
                for key, value in response_dict["extensions"].items()
                if key in KNOWN_RESPONSE_EXTENSIONS
            },
        )

        request = Request(
            method=request_dict["method"],
            url=request_dict["url"],
            headers=[
                (key.encode(HEADERS_ENCODING), value.encode(HEADERS_ENCODING)) for key, value in request_dict["headers"]
            ],
            extensions={
                key: value for key, value in request_dict["extensions"].items() if key in KNOWN_REQUEST_EXTENSIONS
            },
        )

        metadata = Metadata(
            cache_key=metadata_dict["cache_key"],
            created_at=metadata_dict["created_at"],
            number_of_uses=metadata_dict["number_of_uses"],
        )

        return response, request, metadata

    @property
    def is_binary(self) -> bool:  # pragma: no cover
        return False



import typing as tp
from types import TracebackType

import httpcore
import httpx
from httpcore._async.interfaces import AsyncRequestInterface

if tp.TYPE_CHECKING:  # pragma: no cover
    from typing_extensions import Self

__all__ = ("MockAsyncConnectionPool", "MockAsyncTransport")


class MockAsyncConnectionPool(AsyncRequestInterface):
    async def handle_async_request(self, request: httpcore.Request) -> httpcore.Response:
        assert isinstance(request.stream, tp.AsyncIterable)
        data = b"".join([chunk async for chunk in request.stream])  # noqa: F841
        return self.mocked_responses.pop(0)

    def add_responses(self, responses: tp.List[httpcore.Response]) -> None:
        if not hasattr(self, "mocked_responses"):
            self.mocked_responses = []
        self.mocked_responses.extend(responses)

    async def __aenter__(self) -> "Self":
        return self

    async def __aexit__(
        self,
        exc_type: tp.Optional[tp.Type[BaseException]] = None,
        exc_value: tp.Optional[BaseException] = None,
        traceback: tp.Optional[TracebackType] = None,
    ) -> None: ...


class MockAsyncTransport(httpx.AsyncBaseTransport):
    async def handle_async_request(self, request: httpx.Request) -> httpx.Response:
        return self.mocked_responses.pop(0)

    def add_responses(self, responses: tp.List[httpx.Response]) -> None:
        if not hasattr(self, "mocked_responses"):
            self.mocked_responses = []
        self.mocked_responses.extend(responses)
